# -*-coding:utf-8-*-
from .solver import model
from .dao import localDAO
from .logger import Logger


def irt(data_src,
        dao_type='memory',
        theta_bnds=[-4, 4], num_theta=11,
        alpha_bnds=[0.25, 2], beta_bnds=[-2, 2], in_guess_param={},
        model_spec='2PL',
        max_iter=10, tol=1e-3, nargout=2,
        is_parallel=False, num_cpu=6, check_interval=60,
        mode='debug', log_path=None):

    # add logging
    logger = Logger.logger(log_path)

    # load data
    logger.info("start loading data")
    if dao_type == 'memory':
        dao_instance = localDAO(data_src, logger)
    elif dao_type == "db":
        dao_instance = data_src
    else:
        raise ValueError("dao type needs to be either memory or db")
    logger.info("data loaded")

    # setup the model
    if model_spec == '2PL':
        mod = model.IRT_MMLE_2PL(dao_instance,
                logger,
                dao_type=dao_type,
                is_parallel=is_parallel,
                num_cpu=num_cpu,
                check_interval=check_interval,
                mode=mode)
    else:
        raise Exception('Unknown model specification.')

    # specify the irt parameters
    mod.set_options(theta_bnds, num_theta, alpha_bnds, beta_bnds, max_iter, tol)
    mod.set_guess_param(in_guess_param)

    # solve
    mod.solve_EM()
    logger.info("parameter estimated")
    # output
    item_param_dict = mod.get_item_param()
    logger.info("parameter retrieved")

    if nargout == 1:
        return item_param_dict
    elif nargout == 2:
        user_param_dict = mod.get_user_param()
        return item_param_dict, user_param_dict
    else:
        raise Exception('Invalid number of argument')



#!/usr/bin/env python -W ignore
'''
The model is an implementation of EM algorithm of IRT


For reference, see:
Brad Hanson, IRT Parameter Estimation using the EM Algorithm, 2000

The current version only deals with unidimension theta

'''
import numpy as np
from scipy.stats import norm
import time
from copy import deepcopy

from ..util import clib, tools
from ..solver import optimizer
from ..algo import update_theta_distribution

from datetime import datetime
import multiprocessing as mp
from tqdm import tqdm


def procs_operator(procs, TIMEOUT, check_interval):
    for p in procs:
        p.start()

    start = time.time()
    while time.time() - start < TIMEOUT:
        if any(p.is_alive() for p in procs):
            time.sleep(check_interval)
        else:
            for p in procs:
                p.join()
            break
    else:
        for p in procs:
            p.terminate()
            p.join()

        raise Exception('Time out, killing all process')
    return procs


class IRT_MMLE_2PL(object):

    '''
    Exposed methods
    (1) set options
    (2) solve
    (3) get esitmated result
    '''
    def __init__(self,
            dao_instance, logger,
            dao_type='memory',
            is_parallel=False, num_cpu=6, check_interval=60,
            mode='debug'):
        # interface to data
        self.dao = dao_instance
        self.dao_type = dao_type
        self.logger = logger
        self.num_iter = 1
        self.ell_list = []
        self.last_avg_prob = 0

        self.is_parallel = is_parallel
        self.num_cpu = min(num_cpu, mp.cpu_count())
        self.check_interval = check_interval
        self.mode = mode

    def set_options(self, theta_bnds, num_theta, alpha_bnds, beta_bnds, max_iter, tol):
        #  user
        self.num_theta = num_theta
        self._init_user_param(theta_bnds[0], theta_bnds[1], num_theta)
        # item
        boundary = {'alpha': alpha_bnds, 'beta': beta_bnds}
        # solver
        solver_type = 'gradient'
        is_constrained = True

        self._init_solver_param(is_constrained, boundary, solver_type, max_iter, tol)

    def set_guess_param(self, in_guess_param):
        self.guess_param_dict = {}
        for item_idx in range(self.dao.get_num('item')):
            item_id = self.dao.translate('item', item_idx)
            if item_id in in_guess_param:
                self.guess_param_dict[item_idx] = {'c': float(in_guess_param[item_id])}
            else:
                self.guess_param_dict[item_idx] = {'c': 0.0}  # if null then set as 0

    def solve_EM(self):
        # data dependent initialization
        self._init_item_param()

        # main routine
        while True:
            # ----- E step -----
            stime = datetime.now()
            self._exp_step()
            etime = datetime.now()
            runtime = (etime - stime).microseconds / 1000
            self.logger.debug('E step runs for %s sec' % runtime)

            # ----- M step -----
            stime = datetime.now()
            self._max_step()
            etime = datetime.now()
            runtime = (etime - stime).microseconds / 1000
            self.logger.debug('M step runs for %s sec' % runtime)

            # ---- Stop Condition ----
            stime = datetime.now()
            is_stop = self._check_stop()
            etime = datetime.now()
            runtime = (etime - stime).microseconds / 1000
            self.logger.debug('stop condition runs for %s sec' % runtime)

            if is_stop:
                break

    def get_item_param(self):
        output_item_param = {}
        for item_idx in range(self.dao.get_num('item')):
            item_id = self.dao.translate('item', item_idx)
            output_item_param[item_id] = self.item_param_dict[item_idx]
        return output_item_param

    def get_user_param(self):
        output_user_param = {}
        theta_vec = self.__calc_theta()
        for user_idx in range(self.dao.get_num('user')):
            user_id = self.dao.translate('user', user_idx)
            output_user_param[user_id] = theta_vec[user_idx]
        return output_user_param

    '''
    Main Routine
    '''

    def _exp_step(self):
        '''
        Basic Math:
        In the maximization step, need to use E_[j,k](Y=1),E_[j,k](Y=0)
        E(Y=1|param_j,theta_k) = sum_i(data_[i,j]*P(Y=1|param_j,theta_[i,k]))
        since data_[i,j] = 0/1, it is equivalent to sum all done right users

        E(Y=0|param_j,theta_k) = sum_i(
                                (1-data_[i,j]) *(1-P(Y=1|param_j,theta_[i,k])
                                    )
        By similar logic, it is equivalent to sum (1-p) for all done wrong users
        '''

        # (1) update the posterior distribution of theta
        self.__update_theta_distr()

        # (2) marginalize
        # because of the sparsity, the expected right and wrong may not sum up
        # to the total num of items!
        self.__get_expect_count()

    def _max_step(self):
        '''
        Basic Math
            log likelihood(param_j) = sum_k(log likelihood(param_j, theta_k))
        '''
        def update(d, start_idx, end_idx):
            try:
                for item_idx in tqdm(range(start_idx, end_idx)):
                    initial_guess_val = (self.item_param_dict[item_idx]['beta'],
                                        self.item_param_dict[item_idx]['alpha'])
                    opt_worker.set_initial_guess(initial_guess_val)  # the value is a mix of 1/0 and current estimate
                    opt_worker.set_c(self.item_param_dict[item_idx]['c'])

                    # estimate
                    expected_right_count = self.item_expected_right_by_theta[:, item_idx]
                    expected_wrong_count = self.item_expected_wrong_by_theta[:, item_idx]
                    input_data = [expected_right_count, expected_wrong_count]
                    opt_worker.load_res_data(input_data)
                    try:
                        est_param = opt_worker.solve_param_mix(self.is_constrained)
                    except Exception as e:
                        if self.mode == 'production':
                            # In production mode, use the previous iteration
                            self.logger.error('Item %d does not fit' % item_idx)
                            d[item_idx] = self.item_param_dict[item_idx]
                        else:
                            self.logger.critical(e.strerror)
                            raise e
                    finally:
                        d[item_idx] = est_param
            except Exception as e:
                self.logger.critical("Unexpected error:", str(e))
                raise e
        # [A] max for item parameter
        opt_worker = optimizer.irt_2PL_Optimizer()
        # the boundary is universal
        # the boundary is set regardless of the constrained option because the
        # constrained search serves as backup for outlier cases
        opt_worker.set_bounds([self.beta_bnds, self.alpha_bnds])

        # theta value is universal
        opt_worker.set_theta(self.theta_prior_val)
        num_item = self.dao.get_num('item')

        if num_item > self.num_cpu and self.is_parallel:
            num_chunk = self.num_cpu
        else:
            num_chunk = 1

        # [A] calculate p(data,param|theta)
        chunk_list = tools.cut_list(num_item, num_chunk)

        procs = []
        manager = mp.Manager()
        procs_repo = manager.dict()

        for i in range(num_chunk):
            p = mp.Process(target=update, args=(procs_repo, chunk_list[i][0], chunk_list[i][1],))
            procs.append(p)

        if num_chunk > 1:
            procs = procs_operator(procs, 3600, self.check_interval)
        else:
            procs = procs_operator(procs, 7200, 0.1)

        for item_idx in range(num_item):
            self.item_param_dict[item_idx] = {
                    'beta': procs_repo[item_idx][0],
                    'alpha': procs_repo[item_idx][1],
                    'c': self.item_param_dict[item_idx]['c']
                }

        # [B] max for theta density
        self.theta_density = self.posterior_theta_distr.sum(axis=0) / self.posterior_theta_distr.sum()
        self.__check_theta_density()

    def _check_stop(self):
        '''
        preserve user and item parameter from last iteration. This is useful in restoring after a declining llk iteration
        '''
        self.logger.debug('score calculating')
        avg_prob = np.exp(self.__calc_data_likelihood())
        self.logger.debug('score calculated.')

        self.ell_list.append(avg_prob)
        self.logger.debug(avg_prob)

        diff = avg_prob - self.last_avg_prob

        if diff >= 0 and diff <= self.tol:
            self.logger.info('EM converged at iteration %d.' % self.num_iter)
            return True
        elif diff < 0:
            self.item_param_dict = self.last_item_param_dict
            self.logger.info('Likelihood descrease, stops at iteration %d.' % self.num_iter)
            return True
        else:
            # diff larger than tolerance
            # update the stop condition
            self.last_avg_prob = avg_prob
            self.num_iter += 1

            if (self.num_iter > self.max_iter):
                self.logger.info('EM does not converge within max iteration')
                return True
            if self.num_iter != 1:
                self.last_item_param_dict = deepcopy(self.item_param_dict)
            return False

    def _init_solver_param(self, is_constrained, boundary, solver_type, max_iter, tol):
        # initialize bounds
        self.is_constrained = is_constrained
        self.alpha_bnds = boundary['alpha']
        self.beta_bnds = boundary['beta']
        self.solver_type = solver_type
        self.max_iter = max_iter
        self.tol = tol
        if solver_type == 'gradient' and not is_constrained:
            raise Exception('BFGS has to be constrained')

    def _init_item_param(self):
        self.item_param_dict = {}
        for item_idx in range(self.dao.get_num('item')):
            # need to call the old item_id
            c = self.guess_param_dict[item_idx]['c']
            self.item_param_dict[item_idx] = {'alpha': 1.0, 'beta': 0.0, 'c': c}

    def _init_user_param(self, theta_min, theta_max, num_theta, dist='normal'):
        # generte value
        self.theta_prior_val = np.linspace(theta_min, theta_max, num=num_theta)
        if self.num_theta != len(self.theta_prior_val):
            raise Exception('wrong number of inintial theta values')
        # use a normal approximation
        if dist == 'uniform':
            self.theta_density = np.ones(num_theta) / num_theta
        elif dist == 'normal':
            norm_pdf = [norm.pdf(x) for x in self.theta_prior_val]
            normalizer = sum(norm_pdf)
            self.theta_density = np.array([x / normalizer for x in norm_pdf])
        else:
            raise Exception('invalid theta prior distibution %s' % dist)
        self.__check_theta_density()
        # space for each learner
        self.posterior_theta_distr = np.zeros((self.dao.get_num('user'), num_theta))

    def __update_theta_distr(self):
        def update(d, start_idx, end_idx):
            try:
                if self.dao_type == 'db':
                    user2item_conn = self.dao.open_conn('user2item')
                for user_idx in tqdm(range(start_idx, end_idx)):
                    if self.dao_type == 'db':
                        logs = self.dao.get_log(user_idx, user2item_conn)
                    else:
                        logs = self.dao.get_log(user_idx)
                    d[user_idx] = update_theta_distribution(logs, self.num_theta, self.theta_prior_val, self.theta_density, self.item_param_dict)
            except Exception as e:
                self.logger.critical("Unexpected error:", str(e))
                raise e
        # [A] calculate p(data,param|theta)
        num_user = self.dao.get_num('user')
        if num_user > self.num_cpu and self.is_parallel:
            num_chunk = self.num_cpu
        else:
            num_chunk = 1

        # [A] calculate p(data,param|theta)
        chunk_list = tools.cut_list(num_user, num_chunk)
        procs = []
        manager = mp.Manager()
        procs_repo = manager.dict()
        for i in range(num_chunk):
            p = mp.Process(target=update, args=(procs_repo, chunk_list[i][0], chunk_list[i][1],))
            procs.append(p)

        if num_chunk > 1:
            procs = procs_operator(procs, 3600 * 10, self.check_interval)
        else:
            procs = procs_operator(procs, 3600 * 24, 0.1)

        for user_idx in range(num_user):
            self.posterior_theta_distr[user_idx, :] = procs_repo[user_idx]

        # When the loop finish, check if the theta_density adds up to unity for each user
        check_user_distr_marginal = np.sum(self.posterior_theta_distr, axis=1)
        if any(abs(check_user_distr_marginal - 1.0) > 0.0001):
            raise Exception('The posterior distribution of user ability is not proper')

    def __check_theta_density(self):
        if abs(sum(self.theta_density) - 1) > 1e-6:
            raise Exception('theta density does not sum upto 1')

        if self.theta_density.shape != (self.num_theta,):
            raise Exception('theta desnity has wrong shape (%s,%s)' % self.theta_density.shape)

    def __get_expect_count(self):
        def update(d, start_idx, end_idx):
            try:
                if self.dao_type == 'db':
                    item2user_conn = self.dao.open_conn('item2user')
                for item_idx in tqdm(range(start_idx, end_idx)):
                    if self.dao_type == 'db':
                        map_user_idx_vec = self.dao.get_map(item_idx, ['1', '0'], item2user_conn)
                    else:
                        map_user_idx_vec = self.dao.get_map(item_idx, ['1', '0'])
                    d[item_idx] = {
                            1: np.sum(self.posterior_theta_distr[map_user_idx_vec[0], :], axis=0),
                            0: np.sum(self.posterior_theta_distr[map_user_idx_vec[1], :], axis=0)
                            }
            except Exception as e:
                self.logger.critical("Unexpected error:", str(e))
                raise e

        num_item = self.dao.get_num('item')
        if num_item > self.num_cpu and self.is_parallel:
            num_chunk = self.num_cpu
        else:
            num_chunk = 1

        # [A] calculate p(data,param|theta)
        chunk_list = tools.cut_list(num_item, num_chunk)
        procs = []
        manager = mp.Manager()
        procs_repo = manager.dict()
        for i in range(num_chunk):
            p = mp.Process(target=update, args=(procs_repo, chunk_list[i][0], chunk_list[i][1],))
            procs.append(p)

        if num_chunk > 1:
            procs = procs_operator(procs, 5400, self.check_interval)
        else:
            procs = procs_operator(procs, 7200, 0.1)

        self.item_expected_right_by_theta = np.zeros((self.num_theta, self.dao.get_num('item')))
        self.item_expected_wrong_by_theta = np.zeros((self.num_theta, self.dao.get_num('item')))
        for item_idx in range(num_item):
            self.item_expected_right_by_theta[:, item_idx] = procs_repo[item_idx][1]
            self.item_expected_wrong_by_theta[:, item_idx] = procs_repo[item_idx][0]

    def __calc_data_likelihood(self):
        # calculate the likelihood for the data set
        # geometric within learner and across learner
        # 1/N * sum[i](1/Ni *sum[j] (log pij))
        def update(tot_llk, cnt, start_idx, end_idx):
            try:
                if self.dao_type == 'db':
                    user2item_conn = self.dao.open_conn('user2item')
                for user_idx in tqdm(range(start_idx, end_idx)):
                    theta = theta_vec[user_idx]
                    # find all the item_id
                    if self.dao_type == 'db':
                        logs = self.dao.get_log(user_idx, user2item_conn)
                    else:
                        logs = self.dao.get_log(user_idx)
                    if len(logs) == 0:
                        continue
                    ell = 0
                    for log in logs:
                        item_idx = log[0]
                        ans_tag = log[1]
                        alpha = self.item_param_dict[item_idx]['alpha']
                        beta = self.item_param_dict[item_idx]['beta']
                        c = self.item_param_dict[item_idx]['c']
                        ell += clib.log_likelihood_2PL(0.0 + ans_tag, 1.0 - ans_tag, theta, alpha, beta, c)
                    with tot_llk.get_lock():
                        tot_llk.value += ell / len(logs)
                    with cnt.get_lock():
                        cnt.value += 1
            except Exception as e:
                self.logger.critical("Unexpected error:", str(e))
                raise e

        theta_vec = self.__calc_theta()
        num_user = self.dao.get_num('user')
        if num_user > self.num_cpu and self.is_parallel:
            num_chunk = self.num_cpu
        else:
            num_chunk = 1
        user_ell = mp.Value('d', 0.0)
        user_cnt = mp.Value('i', 0)
        chunk_list = tools.cut_list(num_user, num_chunk)
        procs = []
        for i in range(num_chunk):
            p = mp.Process(target=update, args=(user_ell, user_cnt, chunk_list[i][0], chunk_list[i][1],))
            procs.append(p)
        if num_chunk > 1:
            procs = procs_operator(procs, 1200, self.check_interval)
        else:
            procs = procs_operator(procs, 7200, 0.1)
        avg_ell = user_ell.value / user_cnt.value

        return avg_ell

    def __calc_theta(self):
        return np.dot(self.posterior_theta_distr, self.theta_prior_val)



# -*-coding:utf-8-*-

import cython
import numpy as np
cdef extern from "math.h":
    cpdef double exp(double x)
    cpdef double log(double x)

def log_likelihood_2PL(double y1,
                       double y0,
                       double theta,
                       double alpha,
                       double beta,
                       double c=0.0):
    expPos = exp(alpha * theta + beta) ;
    ell =  y1 * log((c + expPos) / (1.0 + expPos)) + y0 * log((1.0 - c) / (1.0 + expPos)) ;

    return ell


def log_likelihood_2PL_gradient(double y1,
                                double y0,
                                double theta,
                                double alpha,
                                double beta,
                                double c=0.0):
    grad = np.zeros(2)

    # It is the gradient of the log likelihood, not the NEGATIVE log likelihood
    temp = exp(beta + alpha * theta)
    beta_grad = temp / (1.0 + temp) * ( y1 * (1.0 - c) / (c + temp) - y0)

    alpha_grad = theta * beta_grad
    grad[0] = beta_grad
    grad[1] = alpha_grad
    return grad



# encoding:utf-8
import logging
import os


class Logger():
    @staticmethod
    def logger(log_path):
        if log_path is not None:
            log_dir = os.path.dirname(log_path)
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            logging.basicConfig(
                level=logging.DEBUG,
                format='%(asctime)s %(pathname)s[line:%(lineno)d] %(levelname)s: %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S',
                filename=log_path,
                filemode='w')
        else:
            logging.basicConfig(
                level=logging.DEBUG,
                format='%(asctime)s %(levelname)s: %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S')

        # 创建一个handler，用于输出到控制台
        console = logging.StreamHandler()
        console.setLevel(logging.DEBUG)

        logger = logging.getLogger()
        # 给logger添加handler
        logger.addHandler(console)
        return logger



'''
This script deals with the data format problem.

The stardard format for pyirt is ( user_id,item_id,result),
where user_id is the idx for test taker, item_id is the idx for items
It is set in this way to deal with the sparsity in the massive dataset.

'''
import numpy as np


def loadFromTuples(data):
    user_ids = []
    item_ids = []
    ans_tags = []
    if len(data) == 0:
        raise Exception('data are empty')

    for log in data:
        user_ids.append(log[0])
        item_ids.append(log[1])
        ans_tags.append(int(log[2]))

    return user_ids, item_ids, ans_tags


def loadFromHandle(fp, sep=','):
    # Default format is comma separated files,
    # Only int is allowed within the environment
    user_ids = []
    item_ids = []
    ans_tags = []

    for line in fp:
        if line == '':
            continue
        user_id_str, item_id_str, ans_tagstr = line.strip().split(sep)
        user_ids.append(user_id_str)
        item_ids.append(item_id_str)
        ans_tags.append(int(ans_tagstr))
    return user_ids, item_ids, ans_tags


def parse_item_paramer(item_param_dict, output_file=None):

    if output_file is not None:
        # open the file
        out_fh = open(output_file, 'w')

    sorted_item_ids = sorted(item_param_dict.keys())

    for item_id in sorted_item_ids:
        param = item_param_dict[item_id]
        alpha_val = np.round(param['alpha'], decimals=2)
        beta_val = np.round(param['beta'], decimals=2)
        if output_file is None:
            print(item_id, alpha_val, beta_val)
        else:
            out_fh.write('{},{},{}\n'.format(item_id, alpha_val, beta_val))


def construct_ref_dict(in_list):
    # map the in_list to a numeric variable from 0 to N
    unique_elements = list(set(in_list))
    element_idxs = range(len(unique_elements))
    idx_ref = dict(zip(unique_elements, element_idxs))
    reverse_idx_ref = dict(zip(element_idxs, unique_elements))
    out_idx_list = [idx_ref[x] for x in in_list]

    return out_idx_list, idx_ref, reverse_idx_ref



# -*- coding:utf-8 -*-
import numpy as np

def irt_fnc(theta, beta, alpha=1.0, c=0.0):
    # beta is item difficulty
    # theta is respondent capability

    prob = c + (1.0 - c) / (1 + np.exp(-(alpha * theta + beta)))
    return prob


def log_likelihood_factor_gradient(y1, y0, theta, alpha, beta, c=0.0):
    temp = np.exp(beta + alpha * theta)
    grad = alpha * temp / (1.0 + temp) * (y1 * (1.0 - c) / (c + temp ) - y0 )

    return grad


def log_likelihood_factor_hessian(y1, y0, theta, alpha, beta, c=0.0):
    x = np.exp(beta + alpha * theta)
    # hessian = - alpha**2*(y1+y0)*temp/(1+temp)**2
    hessian = alpha ** 2 * x / (1 + x) ** 2 * (y1 * (1 - c) * (c - x ** 2) / (c + x) ** 2 - y0)

    return hessian


def log_likelihood_2PL_hessian(y1, y0, theta, alpha, beta, c=0.0):
    hessian = np.zeros((2, 2))
    x = np.exp(beta + alpha * theta)
    base = x / (1 + x) ** 2 * (y1 * (1 - c) * (c - x ** 2) / (c + x) ** 2 - y0)

    hessian = np.matrix([[1, theta], [theta, theta ** 2]]) * base

    return hessian


def logsum(logp):
    w = max(logp)
    logSump = w + np.log(sum(np.exp(logp - w)))
    return logSump



def cut_list(list_length, num_chunk):
    chunk_bnd = [0]
    for i in range(num_chunk):
        chunk_bnd.append(int(list_length*(i+1)/num_chunk))
    chunk_bnd.append(list_length)
    chunk_list = [(chunk_bnd[i], chunk_bnd[i+1]) for i in range(num_chunk) ]
    return chunk_list



# -*- encoding:utf-8 -*-
import io
import time
from collections import defaultdict
import pymongo
from datetime import datetime
import numpy as np
from .util.dao import loadFromHandle, loadFromTuples, construct_ref_dict

from decouple import config
MONGO_USER_NAME = config('MONGO_USER_NAME', default="")
MONGO_PASSWORD = config('MONGO_PASSWORD', default="")
MONGO_ADDRESS = config('MONGO_ADDRESS', default="")
MONGO_AUTH_SOURCE = config('MONGO_AUTH_SOURCE', default="")
MONGO_DB_NAME = config('MONGO_DB_NAME', default="")


class mongoDb(object):
    """ cannot use singleton design, otherwise gets 'Warning: MongoClient opened before fork. Create MongoClient only after forking.'
    """
    def __init__(self):
        mongouri = 'mongodb://{un}:{pw}@{addr}'.format(un=MONGO_USER_NAME, pw=MONGO_PASSWORD, addr=MONGO_ADDRESS)
        if MONGO_AUTH_SOURCE:
            mongouri += '/?authsource={auth_src}'.format(auth_src=MONGO_AUTH_SOURCE)
        # connect
        try:
            self.client = pymongo.MongoClient(mongouri, connect=False, serverSelectionTimeoutMS=10, waitQueueTimeoutMS=100, readPreference='secondaryPreferred')
        except Exception as e:
            raise e
        self.user2item_conn = self.client[MONGO_DB_NAME]['irt_user2item']
        self.item2user_conn = self.client[MONGO_DB_NAME]['irt_item2user']

    def __del__(self):
        self.client.close()


def search_filter(search_id, gid):
    return {'id': search_id, 'gid': gid}


class mongoDAO(object):
    # NOTE: mongoDAO does not use the runtime logger
    # NOTE: The client and the connection is not passed by self because of parallel processing
    def __init__(self, group_id=1, is_msg=False):
        db = mongoDb()
        user_ids = list(set([x['id'] for x in db.user2item_conn.find({'gid': group_id}, {'id': 1})]))
        item_ids = list(set([x['id'] for x in db.item2user_conn.find({'gid': group_id}, {'id': 1})]))

        _, self.user_idx_ref, self.user_reverse_idx_ref = construct_ref_dict(user_ids)
        _, self.item_idx_ref, self.item_reverse_idx_ref = construct_ref_dict(item_ids)

        self.stat = {'user': len(self.user_idx_ref.keys()), 'item': len(self.item_idx_ref.keys())}

        self.gid = group_id
        self.is_msg = is_msg

    def open_conn(self, name):
        if name == "item2user":
            return mongoDb().item2user_conn
        elif name == "user2item":
            return mongoDb().user2item_conn
        else:
            raise ValueError('conn name must be either item2user or user2item')

    def get_num(self, name):
        if name not in ['user', 'item']:
            raise Exception('Unknown stat source %s' % name)
        return self.stat[name]

    def get_log(self, user_idx, user2item_conn):
        user_id = self.translate('user', user_idx)
        # query
        if self.is_msg:
            stime = datetime.now()
            res = user2item_conn.find(search_filter(user_id, self.gid))
            etime = datetime.now()
            search_time = int((etime - stime).microseconds / 1000)
            if search_time > 100:
                print('warning: slow search:%d' % search_time)
        else:
            res = user2item_conn.find(search_filter(user_id, self.gid))
        # parse
        res_num = res.count()
        if res_num == 0:
            return_list = []
        elif res_num > 1:
            raise Exception('duplicate doc for (%s, %d) in user2item' % (user_id, self.gid))
        else:
            log_list = res[0]['data']
            return_list = [(self.item_idx_ref[x[0]], x[1]) for x in log_list]
        return return_list

    def get_map(self, item_idx, ans_key_list, item2user_conn):
        item_id = self.translate('item', item_idx)
        # query
        if self.is_msg:
            stime = datetime.now()
            res = item2user_conn.find(search_filter(item_id, self.gid))
            etime = datetime.now()
            search_time = int((etime - stime).microseconds / 1000)
            if search_time > 100:
                print('warning:slow search:%d' % search_time)
        else:
            res = item2user_conn.find(search_filter(item_id, self.gid))
        # parse
        res_num = res.count()
        if res_num == 0:
            return_list = [[] for ans_key in ans_key_list]
        elif res_num > 1:
            raise Exception('duplicate doc for (%s, %d) in item2user' % (item_id, self.gid))
        else:
            doc = res[0]['data']
            return_list = []
            for ans_key in ans_key_list:
                if str(ans_key) in doc:
                    return_list.append([self.user_idx_ref[x] for x in doc[str(ans_key)]])
                else:
                    return_list.append([])
        return return_list

    def translate(self, data_type, idx):
        if data_type == 'item':
            return self.item_reverse_idx_ref[idx]
        elif data_type == 'user':
            return self.user_reverse_idx_ref[idx]


class localDAO(object):

    def __init__(self, src, logger):

        self.database = localDataBase(src, logger)

        # quasi-bitmap
        user_id_idx_vec, self.user_idx_ref, self.user_reverse_idx_ref = construct_ref_dict(self.database.user_ids)
        item_id_idx_vec, self.item_idx_ref, self.item_reverse_idx_ref = construct_ref_dict(self.database.item_ids)

        self.database.setup(user_id_idx_vec, item_id_idx_vec, self.database.ans_tags)

    def get_num(self, name):
        if name not in ['user', 'item']:
            raise Exception('Unknown stat source %s' % name)
        return self.database.stat[name]

    def get_log(self, user_idx):
        return self.database.user2item[user_idx]

    def get_map(self, item_idx, ans_key_list):
        results = []
        for ans_key in ans_key_list:
            try:
                results.append(self.database.item2user_map[str(ans_key)][item_idx])
            except KeyError:
                results.append([])
        return results

    def close_conn(self):
        pass

    def translate(self, data_type, idx):
        if data_type == 'item':
            return self.item_reverse_idx_ref[idx]
        elif data_type == 'user':
            return self.user_reverse_idx_ref[idx]


class localDataBase(object):
    def __init__(self, src, logger):

        self.logger = logger
        if isinstance(src, io.IOBase):
            # if the src is file handle
            self.user_ids, self.item_ids, self.ans_tags = loadFromHandle(src)
        else:
            # if the src is list of tuples
            self.user_ids, self.item_ids, self.ans_tags = loadFromTuples(src)

    def setup(self, user_idx_vec, item_idx_vec, ans_tags, msg=False):

        start_time = time.time()
        self._process_data(user_idx_vec, item_idx_vec, ans_tags)
        if msg:
            self.logger.debug("--- Process: %f secs ---" % np.round((time.time() - start_time)))

        # initialize some intermediate variables used in the E step
        start_time = time.time()
        self._init_item2user_map()
        if msg:
            self.logger.debug("--- Sparse Mapping: %f secs ---" % np.round((time.time() - start_time)))

    '''
    Need the following dictionary for esitmation routine
    (1) item -> user: key: item_id, value: (user_id, ans_tag)
    (2) user -> item: key: user_id, value: (item_id, ans_tag)
    '''

    def _process_data(self, user_idx_vec, item_idx_vec, ans_tags):
        self.item2user = {}
        self.user2item = defaultdict(list)

        self.stat = {}
        num_log = len(user_idx_vec)
        self.stat['user'] = max(user_idx_vec) + 1  # start count from 0
        self.stat['item'] = max(item_idx_vec) + 1

        for i in range(num_log):
            item_idx = item_idx_vec[i]
            user_idx = user_idx_vec[i]
            ans_tag = ans_tags[i]
            # add to the data dictionary
            if item_idx not in self.item2user:
                self.item2user[item_idx] = defaultdict(list)
            self.item2user[item_idx][ans_tag].append(user_idx)
            self.user2item[user_idx].append((item_idx, ans_tag))

    def _init_item2user_map(self, ans_key_list=['0', '1']):

        self.item2user_map = defaultdict(dict)

        for item_idx, log_result in self.item2user.items():
            for ans_tag, user_idx_vec in log_result.items():
                self.item2user_map[str(ans_tag)][item_idx] = user_idx_vec



# -*- coding:utf-8 -*-
from .util import clib, tools
import numpy as np


def update_theta_distribution(data, num_theta, theta_prior_val, theta_density, item_param_dict):
    '''
    data = [(item_idx int, ans_tag binary)]
    '''

    '''
    Basic Math.
        P_t(theta, data |q_param) = p(data|q_param, theta)*p_[t-1](theta)
        p_t(data|q_param) = sum(p_t(theta,data|q_param)) over theta
        p_t(theta|data, q_param) = P_t(theta, data|q_param)/p_t(data|q_param)
    '''
    likelihood_vec = np.zeros(num_theta)

    for k in range(num_theta):
        theta = theta_prior_val[k]
        ell = 0.0
        for log in data:
            item_idx = log[0]
            ans_tag = log[1]
            alpha = item_param_dict[item_idx]['alpha']
            beta = item_param_dict[item_idx]['beta']
            c = item_param_dict[item_idx]['c']
            ell += clib.log_likelihood_2PL(0.0 + ans_tag, 1.0 - ans_tag, theta, alpha, beta, c)
        likelihood_vec[k] = ell

    # posterior
    joint_llk_vec = likelihood_vec + np.log(theta_density)
    marginal = tools.logsum(joint_llk_vec)
    posterior = np.exp(joint_llk_vec - marginal)

    return posterior



from setuptools import setup

setup(
    name='pyirt',
    version="0.3.4",
    packages=['pyirt',
              'pyirt/solver',
              'pyirt/util', ],
    license='MIT',
    description='A python implementation of Item Response Theory(IRT), specializing in big dataset',
    author='Junchen Feng',
    author_email='frankfeng.pku@gmail.com',
    include_package_data=True,
    url='https://github.com/junchenfeng/pyirt',
    download_url='https://github.com/17zuoye/pyirt/archive/v0.3.4.tar.gz',
    keywords=['IRT', 'EM', 'big data'],
    zip_safe=False,
    platforms='any',
    install_requires=['numpy',
                      'scipy',
                      'cython',
                      'pymongo',
                      'tqdm',
                      'python-decouple'],

    package_data={'pyirt': ["*.pyx"]},

    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
    ],
)
